@@ -3,23 +3,30 @@ NULL
 
 #' Odbc Connection Methods
 #'
-#' Implementations of pure virtual functions defined in the \code{DBI} package
+#' Implementations of pure virtual functions defined in the `DBI` package
 #' for OdbcConnection objects.
 #' @name OdbcConnection
 NULL
 
-OdbcConnection <- function(dsn = NULL, ..., driver = NULL, server = NULL, database = NULL, uid = NULL, pwd = NULL, .connection_string = NULL) {
+class_cache <- new.env(parent = emptyenv())
+
+OdbcConnection <- function(dsn = NULL, ..., encoding = "", timezone = "UTC", driver = NULL, server = NULL, database = NULL, uid = NULL, pwd = NULL, .connection_string = NULL) {
   args <- c(dsn = dsn, driver = driver, server = server, database = database, uid = uid, pwd = pwd, list(...))
   stopifnot(all(has_names(args)))
 
   connection_string <- paste0(.connection_string, paste(collapse = ";", sep = "=", names(args), args))
-  ptr <- odbc_connect(connection_string)
+  ptr <- odbc_connect(connection_string, timezone = timezone, encoding = encoding)
   quote <- connection_quote(ptr)
 
   info <- connection_info(ptr)
-  class(info) <- c(info$dbms.name, "list")
+  class(info) <- c(info$dbms.name, "driver_info", "list")
 
-  new("OdbcConnection", ptr = ptr, quote = quote, info = info)
+  class <- getClassDef(info$dbms.name, where = class_cache, inherits = FALSE)
+  if (is.null(class)) {
+    setClass(info$dbms.name,
+      contains = "OdbcConnection", where = class_cache)
+  }
+  new(info$dbms.name, ptr = ptr, quote = quote, info = info, encoding = encoding)
 }
 
 #' @rdname OdbcConnection
@@ -30,7 +37,8 @@ setClass(
   slots = list(
     ptr = "externalptr",
     quote = "character",
-    info = "ANY"
+    info = "ANY",
+    encoding = "character"
   )
 )
 
@@ -81,6 +89,7 @@ setMethod(
       warning("Connection already closed.", call. = FALSE)
     }
 
+    on_connection_closed(conn)
     connection_release(conn@ptr)
     invisible(TRUE)
   })
@@ -102,6 +111,7 @@ setMethod(
   "dbSendStatement", c("OdbcConnection", "character"),
   function(conn, statement, ...) {
     res <- OdbcResult(connection = conn, statement = statement)
+    result_execute(res@ptr)
     res
   })
 
@@ -111,7 +121,16 @@ setMethod(
 setMethod(
   "dbDataType", "OdbcConnection",
   function(dbObj, obj, ...) {
-    get_data_type(dbObj@info, obj)
+    odbcDataType(dbObj, obj)
+  })
+
+#' @rdname OdbcConnection
+#' @inheritParams DBI::dbDataType
+#' @export
+setMethod(
+  "dbDataType", c("OdbcConnection", "data.frame"),
+  function(dbObj, obj, ...) {
+    vapply(obj, odbcDataType, con = dbObj, FUN.VALUE = character(1), USE.NAMES = TRUE)
   })
 
 #' @rdname OdbcConnection
@@ -136,6 +155,41 @@ setMethod(
     DBI::SQL(paste(conn@quote, encodeString(x), conn@quote, sep = ""))
   })
 
+
+#' Un-Quote identifiers
+#'
+#' Call this method to generate a string that is unquoted. This is the inverse
+#' of `DBI::dbQuoteIdentifier`.
+#'
+#' @param x A character vector to un-quote.
+#' @inheritParams DBI::dbQuoteIdentifier
+#' @export
+setGeneric(
+  "dbUnQuoteIdentifier",
+  function(conn, x, ...) standardGeneric("dbUnQuoteIdentifier")
+)
+
+#' @rdname dbUnQuoteIdentifier
+#' @inheritParams DBI::dbQuoteIdentifier
+#' @export
+setMethod(
+  "dbUnQuoteIdentifier", c("OdbcConnection", "SQL"),
+  function(conn, x) {
+    x <- as.character(x)
+    x <- gsub(paste0("^", conn@quote), "", x)
+    x <- gsub(paste0(conn@quote, "$"), "", x)
+    x
+  })
+
+#' @rdname dbUnQuoteIdentifier
+#' @inheritParams DBI::dbQuoteIdentifier
+#' @export
+setMethod(
+  "dbUnQuoteIdentifier", c("OdbcConnection", "character"),
+  function(conn, x) {
+    x
+  })
+
 #' @rdname OdbcConnection
 #' @inheritParams DBI::dbListTables
 #' @export
@@ -151,7 +205,8 @@ setMethod(
 setMethod(
   "dbExistsTable", c("OdbcConnection", "character"),
   function(conn, name, ...) {
-    name %in% dbListTables(conn)
+    stopifnot(length(name) == 1)
+    dbUnQuoteIdentifier(conn, name) %in% dbListTables(conn)
   })
 
 #' @rdname OdbcConnection
@@ -170,7 +225,8 @@ setMethod(
   "dbRemoveTable", c("OdbcConnection", "character"),
   function(conn, name, ...) {
     name <- dbQuoteIdentifier(conn, name)
-    dbGetQuery(conn, paste("DROP TABLE ", name))
+    dbExecute(conn, paste("DROP TABLE ", name))
+    on_connection_updated(conn, name)
     invisible(TRUE)
   })
 
@@ -180,18 +236,20 @@ setMethod(
 setMethod(
   "dbGetInfo", "OdbcConnection",
   function(dbObj, ...) {
-    connection_info(dbObj@ptr)
+    info <- connection_info(dbObj@ptr)
+    structure(info, class = c(info$dbms.name, "driver_info", "list"))
   })
 
 #' @rdname OdbcConnection
 #' @inheritParams DBI::dbGetQuery
+#' @inheritParams DBI::dbFetch
 #' @export
 setMethod("dbGetQuery", signature("OdbcConnection", "character"),
-  function(conn, statement, ...) {
+  function(conn, statement, n = -1, ...) {
     rs <- dbSendQuery(conn, statement, ...)
     on.exit(dbClearResult(rs))
 
-    df <- dbFetch(rs, n = -1, ...)
+    df <- dbFetch(rs, n = n, ...)
 
     if (!dbHasCompleted(rs)) {
       warning("Pending rows", call. = FALSE)
@@ -231,112 +289,22 @@ setMethod(
     invisible(TRUE)
   })
 
-get_data_type <- function(info, obj, ...) UseMethod("get_data_type")
-
-varchar <- function(x, type = "varchar") {
-  max_length <- max(nchar(as.character(x)), na.rm = TRUE)
-  paste0(type, "(", max(255, max_length), ")")
-}
-
-varbinary <- function(x, type = "varbinary") {
-  max_length <- max(lengths(x), na.rm = TRUE)
-  paste0(type, "(", max(255, max_length), ")")
-}
-
-get_data_type.default <- function(info, obj, ...) {
-  if (is.factor(obj)) return(varchar(obj))
-  if (is(obj, "POSIXct")) return("TIMESTAMP")
-  if (is(obj, "Date")) return("DATE")
-  if (is(obj, "blob")) return("BLOB")
-
-  switch(typeof(obj),
-    integer = "INTEGER",
-    double = "DOUBLE PRECISION",
-    character = varchar(obj),
-    logical = "SMALLINT",
-    list = varchar(obj),
-    stop("Unsupported type", call. = FALSE)
-  )
-}
-
-get_data_type.MySQL <- function(info, obj, ...) {
-  if (is.factor(obj)) return("TEXT")
-  if (is(obj, "POSIXct")) return("DATETIME")
-  if (is(obj, "Date")) return("DATE")
-  if (is(obj, "blob")) return("BLOB")
-
-  switch(typeof(obj),
-    integer = "INTEGER",
-    double = "DOUBLE",
-    character = "TEXT",
-    logical = "TINYINT",
-    list = "TEXT",
-    stop("Unsupported type", call. = FALSE)
-  )
-}
-
-get_data_type.PostgreSQL <- function(info, obj, ...) {
-  if (is.factor(obj)) return("TEXT")
-  if (is(obj, "POSIXct")) return("TIMESTAMP")
-  if (is(obj, "Date")) return("DATE")
-  if (is(obj, "blob")) return("bytea")
-
-  switch(typeof(obj),
-    integer = "INTEGER",
-    double = "DOUBLE PRECISION",
-    character = "TEXT",
-    logical = "BOOLEAN",
-    list = "TEXT",
-    stop("Unsupported type", call. = FALSE)
-  )
-}
-
-`get_data_type.Microsoft SQL Server` <- function(info, obj, ...) {
-  if (is.factor(obj)) return(varchar(obj))
-  if (is(obj, "POSIXct")) return("datetime")
-  if (is(obj, "Date")) return("date")
-  if (is(obj, "blob")) return(varbinary(obj))
-
-  switch(typeof(obj),
-    integer = "int",
-    double = "float",
-    character = varchar(obj),
-    logical = "BIT",
-    list = varchar(obj),
-    stop("Unsupported type", call. = FALSE)
-  )
-}
-
-`get_data_type.SQLite` <- function(info, obj, ...) {
-  if (is.factor(obj)) return("TEXT")
-  if (is(obj, "POSIXct")) return("NUMERIC")
-  if (is(obj, "Date")) return("NUMERIC")
-  if (is(obj, "blob")) return("BLOB")
-
-  switch(typeof(obj),
-    integer = "INTEGER",
-    double = "REAL",
-    character = "TEXT",
-    logical = "NUMERIC",
-    list = "TEXT",
-    stop("Unsupported type", call. = FALSE)
-  )
-}
-
 #' List Available ODBC Drivers
 #'
 #' @return A data frame with three columns.
 #' If a given driver does not have any attributes the last two columns will be
-#' \code{NA}.
+#' `NA`.
 #' \describe{
 #'   \item{name}{Name of the driver}
 #'   \item{attribute}{Driver attribute name}
 #'   \item{value}{Driver attribute value}
 #' }
 #' @export
-list_drivers <- function() {
+odbcListDrivers <- function() {
   res <- list_drivers_()
-  res[res == ""] <- NA_character_
+  if (nrow(res) > 0) {
+    res[res == ""] <- NA_character_
+  }
   res
 }
 
@@ -348,6 +316,31 @@ list_drivers <- function() {
 #'   \item{description}{Data Source description}
 #' }
 #' @export
-list_data_sources <- function() {
+odbcListDataSources <- function() {
   list_data_sources_()
 }
+
+#' Set the Transaction Isolation Level for a Connection
+#'
+#' @param levels One or more of \Sexpr[stage=render, results=rd]{odbc:::choices_rd(names(odbc:::transactionLevels()))}.
+#' @inheritParams DBI::dbDisconnect
+#' @seealso \url{https://docs.microsoft.com/en-us/sql/odbc/reference/develop-app/setting-the-transaction-isolation-level}
+#' @export
+#' @noMd
+#' @examples
+#' \dontrun{
+#'   # Can use spaces or underscores in between words.
+#'   odbcSetTransactionIsolationLevel(con, "read uncommitted")
+#'
+#'   # Can also use the full constant name.
+#'   odbcSetTransactionIsolationLevel(con, "SQL_TXN_READ_UNCOMMITTED")
+#' }
+odbcSetTransactionIsolationLevel <- function(conn, levels) {
+  # Convert to lowercase, spaces to underscores, remove sql_txn prefix
+  levels <- tolower(levels)
+  levels <- gsub(" ", "_", levels)
+  levels <- sub("sql_txn_", "", levels)
+  levels <- match.arg(tolower(levels), names(transactionLevels()), several.ok = TRUE)
+
+  set_transaction_isolation(conn@ptr, transactionLevels()[levels])
+}
